library(foreign)
library(Hmisc)
library(survey)
library(xtable)      
library(np)

wtd.quantile <- function (x, weights = NULL, probs = c(0, 0.25, 0.5, 0.75, 1),
    type = c("quantile", "(i-1)/(n-1)", "i/(n+1)", "i/n"), normwt = FALSE,
    na.rm = TRUE)
{
    if (!length(weights))
        return(quantile(x, probs = probs, na.rm = na.rm))
    type <- match.arg(type)
    if (any(probs < 0 | probs > 1))
        stop("Probabilities must be between 0 and 1 inclusive")
    nams <- paste(format(round(probs * 100, if (length(probs) >
        1) 2 - log10(diff(range(probs))) else 2)), "%", sep = "")
    if (type == "quantile") {
        w <- wtd.table(x, weights, na.rm = na.rm, normwt = normwt,
            type = "list")
        x <- w$x
        wts <- w$sum.of.weights
        n <- sum(wts)
        order <- 1 + (n - 1) * probs
        low <- pmax(floor(order), 1)
        high <- pmin(low + 1, n)
        order <- order%%1
        allq <- approx(cumsum(wts), x, xout = c(low, high), method = "constant",
            f = 1, rule = 2)$y
        k <- length(probs)
        quantiles <- (1 - order) * allq[1:k] + order * allq[-(1:k)]
        names(quantiles) <- nams
        return(quantiles)
    }
    w <- wtd.Ecdf(x, weights, na.rm = na.rm, type = type, normwt = normwt)
    structure(approx(w$ecdf, w$x, xout = probs, rule = 2, f=1,method = "constant",ties="ordered")$y,
        names = nams)
}




fppe.para <- function(formula, discrete, dat.base, dat.treat, weights.base = NA, weights.treat = NA, bs = FALSE, bsrep=200, y.grid = 50, y.approx=1000) {

   n.base <- dim(dat.base)[1]
   n.treat <- dim(dat.treat)[1]

   # check for presence of weights. if false assign 1/n
   if (sum(is.na(weights.base))>0) {weights.base <- rep(1,n.base)}
   if (sum(is.na(weights.treat))>0) {weights.treat <- rep(1,n.treat)}

   # get variable names
   outname <- all.vars(formula)[1]
   covname <- all.vars(formula)[-1]
   discname <- all.vars(discrete)

   # determine y.grid
   y.grid <- wtd.quantile(dat.base[,outname],seq(0,1,length=y.grid),w=weights.base, normwt=TRUE,type="i/n")
   y.grid2 <- quantile(c(dat.base[,outname],dat.treat[,outname]),seq(0,1,length=y.approx))


 computecdf <- function(dat.base, dat.treat, weights.base, weights.treat,...) {

   n.base <- dim(dat.base)[1]
   n.treat <- dim(dat.treat)[1]
  
   # copy outcome variable to other colum
   dat.base$eval <- NA
   dat.base[,"eval"] <- dat.base[,outname]

   # get type of variable and construct transformation
   dat.comp <- list(NA)
   dat.comp.alt <- list(NA) # only used in partially identified case

   Fw <- matrix(NA,ncol=length(covname), nrow=n.base)
   Fw.minus <- matrix(NA,ncol=length(covname), nrow=n.base)

   for (j in 1:length(covname)) {

    if(covname[j] %in% discname) {

     dat.comp[[j]] <- list(NA)
     dat.comp[[j]][[1]] <- dat.base[,covname[j]]
     dat.comp[[j]][[2]] <- dat.base[,covname[j]]

     wecdf <- wtd.Ecdf(dat.base[,covname[j]], w=weights.base, normwt=TRUE)

     Fw[,j] <- approx(wecdf$x[-1], wecdf$ecdf[-1], xout = dat.base[,covname[j]],  f=1, yleft=0, yright=1, method = "constant")$y
     Fw.minus[,j] <- approx(wecdf$x[-1], c(0,wecdf$ecdf[-c(1,length(wecdf$ecdf))]), xout = dat.base[,covname[j]], f=1, yleft=0, yright=1, method = "constant")$y

     dat.comp[[j]][[1]] <- (wtd.quantile(dat.treat[,covname[j]],Fw[,j],w=weights.treat, normwt=TRUE,type="i/n"))
     dat.comp[[j]][[2]] <- (wtd.quantile(dat.treat[,covname[j]],Fw.minus[,j],w=weights.treat, normwt=TRUE,type="i/n"))

    } else {

  dat.comp[[j]] <- dat.base[,covname[j]]

  tiebrake <- runif(length(dat.base[,covname[j]]),0,.0001)

  wecdf <- wtd.Ecdf(dat.base[,covname[j]] +tiebrake, w=weights.base, normwt=TRUE)
  wecdf <- approx(wecdf$x[-1], wecdf$ecdf[-1], xout = dat.base[,covname[j]]+tiebrake, f=1, yleft=0, yright=1, method = "constant")$y

  dat.comp[[j]] <- wtd.quantile(dat.treat[,covname[j]],wecdf,w=weights.treat, normwt=TRUE,type="i/n")

 }
}

# estimate conditional CDFs using formula; get counterfactual cdf


 ###############################################################################
 # BEGIN: initialize the CDF vector

 counter.cdf <- list(NA)

 for (j in 1:length(covname)) {
  if(covname[j] %in% discname) {
   counter.cdf[[j]] <- list(NA)
   counter.cdf[[j]][[1]] <- c(0,rep(1, length(y.grid)-1))
   counter.cdf[[j]][[2]] <- c(0,rep(1, length(y.grid)-1))
  } else {
   counter.cdf[[j]] <- c(0,rep(1, length(y.grid)-1))
  }
 }
 counter.cdf[[j+1]] <- c(0,rep(1, length(y.grid)-1))
 # END: initialize the CDF vector
 ###############################################################################

# loop through all grid points

# set progress bar
pb <- txtProgressBar(style = 3)

for (i in 2:(length(y.grid)-1)) {

 dat.base[,outname] <-  I(dat.base[,"eval"] <= y.grid[i])
 design <- svydesign(~0, data=dat.base, weights=~weights.base)
 mod <- svyglm(formula, data=dat.base, design=design, family=quasibinomial(link="logit"), start=(if (exists("mod")) {coef(mod)} ),
               control=list(epsilon=1e-8,maxit=25,trace=FALSE))

 # loop over all variables
 for (j in 1:length(covname)) {

  if (covname[j] %in% discname) {

   # Generate rank allocators and corresponding data
   dat.temp1 <- dat.base
   dat.temp1[,covname[j]] <- dat.comp[[j]][[1]]
   dat.temp2 <- dat.base
   dat.temp2[,covname[j]] <- dat.comp[[j]][[2]]

   cdfdiff <- predict(mod, newdata=dat.temp1, type="response",se=FALSE) - predict(mod, newdata=dat.temp2, type="response",se=FALSE)

   w.vals <- unique(dat.base[,covname[j]])
   v.u <- rep(0,n.base)
   v.l <- rep(0,n.base)

   # loop over all support points
   for (k in 1:length(w.vals)) {
     sel <- which(dat.base[,covname[j]] == w.vals[k])
     wecdf <- wtd.Ecdf(cdfdiff[sel], w=weights.base[sel], normwt=TRUE)
     wecdf <- approx(wecdf$x[-1], wecdf$ecdf[-1], xout = cdfdiff[sel], f=1, yleft=0, yright=1, method = "constant")$y
     v.u[sel] <- wecdf
     v.l[sel] <- 1-wecdf
   }
   dat.temp1[,covname[j]] <- (wtd.quantile(dat.treat[,covname[j]],Fw.minus[,j] + v.u*(Fw[,j]-Fw.minus[,j]),w=weights.treat, normwt=TRUE,type="i/n"))
   dat.temp2[,covname[j]] <- (wtd.quantile(dat.treat[,covname[j]],Fw.minus[,j] + v.l*(Fw[,j]-Fw.minus[,j]),w=weights.treat, normwt=TRUE,type="i/n"))

   counter.cdf[[j]][[1]][i] <- weighted.mean(predict(mod,newdata=dat.temp1, type="response"), w =  weights.base)
   counter.cdf[[j]][[2]][i] <- weighted.mean(predict(mod,newdata=dat.temp2, type="response"), w =  weights.base)

  rm(dat.temp1,dat.temp2)

  } else {
   dat.temp <- dat.base
   dat.temp[,covname[j]] <- dat.comp[[j]]
   counter.cdf[[j]][i] <- weighted.mean(predict(mod,newdata=dat.temp, type="response"), w =  weights.base)
   rm(dat.temp)
  }
 } # END LOOP OVER VARIABLES

 counter.cdf[[j+1]][i] <- weighted.mean(predict(mod,newdata=dat.treat, type="response"), w =  weights.treat)

 setTxtProgressBar(pb, i/(length(y.grid)-2))


} # END LOOP OVER GRID POINTS
close(pb)     # close progress bar

 dat.base[,outname] <-  dat.base[,"eval"]


###############################################################################
# Begin: Obtain point probabilities

 probs <- list(NA)

 # BEGIN: Loop over all variables
 for (j in 1:length(covname)) {
  if(covname[j] %in% discname) {
   probs[[j]] <- list(NA)
   probs[[j]][[1]] <- diff(approx(x=y.grid, y=sort(counter.cdf[[j]][[1]]), xout=y.grid2, rule=2)$y)
   probs[[j]][[2]] <- diff(approx(x=y.grid, y=sort(counter.cdf[[j]][[2]]), xout=y.grid2, rule=2)$y)
  } else {
   probs[[j]] <- diff(approx(x=y.grid, y=sort(counter.cdf[[j]]), xout=y.grid2, rule=2)$y)
  }
 }
 # END LOOP OVER VARIABLES
 probs[[j+1]] <- diff(approx(x=y.grid, y=sort(counter.cdf[[j+1]]), xout=y.grid2, rule=2 )$y)
 
 cdfbase <- wtd.Ecdf(dat.base[,outname], w=weights.base, normwt=TRUE) 
 probs[[j+2]] <- diff(approx(x=cdfbase$x, y=cdfbase$ecdf, xout=y.grid2, rule=2 )$y) # base probabilities
 probs[[j+2]] <-  probs[[j+2]]/sum( probs[[j+2]])
 
 cdftreat <- wtd.Ecdf(dat.treat[,outname], w=weights.treat, normwt=TRUE) 
 probs[[j+3]] <- diff(approx(x=cdftreat$x, y=cdftreat$ecdf, xout=y.grid2, rule=2)$y) # treat probabilities
 probs[[j+3]] <- probs[[j+3]]/sum( probs[[j+3]])
 names(probs) <- c(covname,"full","base","treat")

# END: Obtain point probabilities
###############################################################################

  output <- list(y.grid = y.grid2[-y.approx] + diff(y.grid2)/2,
                 covname = covname,
                 discname = discname,          
                 probs = probs)
  output
 }

 output <- computecdf(dat.base, dat.treat, weights.base, weights.treat, y.grid, y.grid2)

 output.bs <- list(NA)
  
 if  (bs == TRUE) {
  bsskip <- 1
  for (i in 1:bsrep) {
    bsamp1 <- sample(1:n.base, round(n.base/bsskip), replace=TRUE, prob=weights.base)
    bsamp2 <- sample(1:n.treat, round(n.treat/bsskip), replace=TRUE, prob=weights.treat)
    print(c("Bootstrap Replication #:",i))
    output.bs[[i]] <- computecdf(dat.base[bsamp1,], dat.treat[bsamp2,], weights.base=rep(1,length(bsamp1)), weights.treat=rep(1,length(bsamp2)), y.grid, y.grid2 )
  }
 }
 output <-  list(estimate = output, bs = output.bs)
 output

}

###################################################################################
fppe.quantile.proc <- function(output, qseq, ...) {

 base <- wtd.quantile(output$y.grid,w=output$probs$base,qseq,normwt=TRUE,type="i/n")
 treat <- wtd.quantile(output$y.grid,w=output$probs$treat,qseq,normwt=TRUE,type="i/n")

 quant <- list(NA)

 # BEGIN: Loop over all variables
 for (j in 1:length(output$covname)) {

  if(output$covname[j] %in% output$discname) {
   quant[[j]] <- list(NA)
   quant[[j]][[1]] <- wtd.quantile(output$y.grid,w=output$probs[[j]][[1]],qseq,normwt=TRUE,type="i/n") - base
   quant[[j]][[2]] <- wtd.quantile(output$y.grid,w=output$probs[[j]][[2]],qseq,normwt=TRUE,type="i/n") - base
  } else {
   quant[[j]] <- wtd.quantile(output$y.grid,w=output$probs[[j]],qseq,normwt=TRUE,type="i/n") - base
  }

 }
 # END LOOP OVER VARIABLES

 quant[[j+1]] <- wtd.quantile(output$y.grid,w=output$probs[[j+1]],qseq,normwt=TRUE,type="i/n") - base# Full Policy Effect
 quant[[j+2]] <- treat - base  # Observed change
 names(quant) <- c(output$covname,"full","observed")

 out <- list(quant = quant, qseq = qseq)
}



fppe.mean.proc <- function(output, ...) {

 base <- wtd.mean(output$y.grid,w=output$probs$base,normwt=TRUE)
 treat <- wtd.mean(output$y.grid,w=output$probs$treat,normwt=TRUE)

 meanval <- list(NA)

 # BEGIN: Loop over all variables
 for (j in 1:length(output$covname)) {

  if(output$covname[j] %in% output$discname) {
   meanval[[j]] <- list(NA)
   meanval[[j]][[1]] <- wtd.mean(output$y.grid,w=output$probs[[j]][[1]],normwt=TRUE) - base
   meanval[[j]][[2]] <- wtd.mean(output$y.grid,w=output$probs[[j]][[2]],normwt=TRUE) - base
  } else {
   meanval[[j]] <- wtd.mean(output$y.grid,w=output$probs[[j]],normwt=TRUE) - base
  }

 }
 # END LOOP OVER VARIABLES

 meanval[[j+1]] <- wtd.mean(output$y.grid,w=output$probs[[j+1]],normwt=TRUE) - base             # Full Policy Effect
 meanval[[j+2]] <- treat - base  # Observed change

 names(meanval) <- c(output$covname,"full","observed")

 out <- list(mean=meanval)


 # END: Mean
 ###############################################################################

 out

}




fppe.var.proc <- function(output, mu.vals, varmax=FALSE ,...) {

compressed <- function(mu,f.l,f.u,ev.points) {
 k <- 1
 while (wtd.mean(ev.points,w=diff(c(0,sort(c(f.u[1:k],f.l[(k+1):length(ev.points)] ))))) < mu) {
  k <- k+1
  if (k == (length(ev.points)-1)) break
 }
 diff(c(0,sort(c(f.u[1:k],f.l[(k+1):length(ev.points)] ))))
}

dispersed <- function(mu,f.l,f.u,ev.points) {
 a <- round(min(f.l),2)
 repeat {
  a <- a + .01
  ww <- diff(c(0,sort(c(f.l[f.l < a] ,rep(a,length(ev.points)-sum(f.l < a)-sum(f.u >= a)),f.u[f.u >= a] ))))
  if (wtd.mean(ev.points,w=ww)<mu) break
  if (round(a,2)==.99 )  break
  }
  ww
}


 base <- wtd.var(output$y.grid,w=output$probs$base,normwt=TRUE)
 treat <- wtd.var(output$y.grid,w=output$probs$treat,normwt=TRUE)

 mean.base <- wtd.mean(output$y.grid,w=output$probs$base,normwt=TRUE)

 meanval <- list(NA)

 # BEGIN: Loop over all variables
 for (j in 1:length(output$covname)) {

  if(output$covname[j] %in% output$discname) {
   meanval[[j]] <- list(NA)
   meanval[[j]][[1]] <- rep(0,10)
   meanval[[j]][[2]] <- rep(0,10)

   mu.eval <- seq(from = mu.vals[[1]][[j]][[1]], to = mu.vals[[1]][[j]][[2]], length=12)[-c(1,12)] + mean.base

   for (k in 1:10) {
   meanval[[j]][[1]][k] <- wtd.var(output$y.grid,w=compressed(mu.eval[k],cumsum(output$probs[[j]][[1]]),cumsum(output$probs[[j]][[2]]),output$y.grid),normwt=TRUE) - base
   meanval[[j]][[2]][k] <- wtd.var(output$y.grid,w=dispersed(mu.eval[k],cumsum(output$probs[[j]][[1]]),cumsum(output$probs[[j]][[2]]),output$y.grid),normwt=TRUE) - base
   }
   
   if (varmax==TRUE) {
    meanval[[j]][[1]] <- max(meanval[[j]][[1]])
    meanval[[j]][[2]] <- min(meanval[[j]][[2]])
   }
   
  } else {
   meanval[[j]] <- wtd.var(output$y.grid,w=output$probs[[j]],normwt=TRUE) - base
  }

 }
 # END LOOP OVER VARIABLES

 meanval[[j+1]] <- wtd.var(output$y.grid,w=output$probs[[j+1]],normwt=TRUE) - base    # Full Policy Effect
 meanval[[j+2]] <- treat - base  # Observed change

 names(meanval) <- c(output$covname,"full","observed")

 out <- list(var=meanval, mean=mu.vals[[1]])


 # END: VAR
 ###############################################################################

 out

}



fppe.qc.proc <- function(output, qc = c(.1,.9) ,...) {

 base <- diff(wtd.quantile(output$y.grid,w=output$probs$base,qc,normwt=TRUE,type="i/n"))
 treat <- diff(wtd.quantile(output$y.grid,w=output$probs$treat,qc,normwt=TRUE,type="i/n"))

 meanval <- list(NA)

 # BEGIN: Loop over all variables
 for (j in 1:length(output$covname)) {

  if (output$covname[j] %in% output$discname) {
  
   meanval[[j]] <- list(NA)

   g <- mean(qc) # must be in qc
   m <- mean( c(wtd.quantile(output$y.grid,w=output$probs[[j]][[2]],g,normwt=TRUE,type="i/n"),
                wtd.quantile(output$y.grid,w=output$probs[[j]][[1]],g,normwt=TRUE,type="i/n")))
   ww.comp <- diff(c(0,sort(c(cumsum(output$probs[[j]][[2]])[ output$y.grid < m ], cumsum(output$probs[[j]][[1]])[output$y.grid >= m ] ))))
   ww.disp <- diff(c(0,sort(c(cumsum(output$probs[[j]][[1]])[cumsum(output$probs[[j]][[1]]) < g] ,rep(g,length(output$y.grid)-sum(cumsum(output$probs[[j]][[1]]) < g)-sum(cumsum(output$probs[[j]][[2]]) >= g)),cumsum(output$probs[[j]][[2]])[cumsum(output$probs[[j]][[2]]) >= g] ))))

   meanval[[j]][[1]] <- diff(wtd.quantile(output$y.grid,w=ww.comp,qc,normwt=TRUE,type="i/n")) - base
   meanval[[j]][[2]] <- diff(wtd.quantile(output$y.grid,w=ww.disp,qc,normwt=TRUE,type="i/n")) - base
   
   names(meanval[[j]][[1]]) <- NULL
   names(meanval[[j]][[2]]) <- NULL

  } else {
   meanval[[j]] <- diff(wtd.quantile(output$y.grid,w=output$probs[[j]],qc,normwt=TRUE,type="i/n"))- base
   names(meanval[[j]]) <- NULL
  }

 }
 # END LOOP OVER VARIABLES

 meanval[[j+1]] <- diff(wtd.quantile(output$y.grid,w=output$probs[[j+1]],qc,normwt=TRUE,type="i/n"))- base # Full Policy Effect
 meanval[[j+2]] <- treat - base  # Observed change
 names(meanval[[j+1]]) <- NULL
 names(meanval[[j+2]]) <- NULL
 names(meanval) <- c(output$covname,"full","observed")

 out <- list(qc=meanval,contrast=qc)


 # END: QC
 ###############################################################################

 out

}



fppe.gini.proc <- function(output, mu.vals, ginimax=FALSE ,...) {

 compressed <- function(mu,f.l,f.u,ev.points) {
  k <- 1
  while (wtd.mean(ev.points,w=diff(c(0,sort(c(f.u[1:k],f.l[(k+1):length(ev.points)] ))))) < mu) {
   k <- k+1
   if (k == length(ev.points)) break
  }
  diff(c(0,sort(c(f.u[1:k],f.l[(k+1):length(ev.points)] ))))
 }

 dispersed <- function(mu,f.l,f.u,ev.points) {
  a <- 0
  repeat {
   a <- a + .01
   ww <- diff(c(0,sort(c(f.l[f.l < a] ,rep(a,length(ev.points)-sum(f.l < a)-sum(f.u >= a)),f.u[f.u >= a] ))))
   if (wtd.mean(ev.points,w=ww)<mu) break
   if (round(a,2)==.99 )  break
  }
  ww
 }

 Gini <- function (x, w) {
    n <- length(x)
    w <- w/sum(w)
    G <- sum(x[order(x)] * 1:n * w[order(x)])
    G <- 2 * G/(n*sum(x[order(x)]  * w[order(x)]))
    G - 1 - (1/n)
 }



 base <- Gini(output$y.grid,w=output$probs$base)
 treat <-Gini(output$y.grid,w=output$probs$treat)

 mean.base <- wtd.mean(output$y.grid,w=output$probs$base,normwt=TRUE)

 meanval <- list(NA)

 # BEGIN: Loop over all variables
 for (j in 1:length(output$covname)) {

  if(output$covname[j] %in% output$discname) {
   meanval[[j]] <- list(NA)
   meanval[[j]][[1]] <- rep(0,10)
   meanval[[j]][[2]] <- rep(0,10)

   mu.eval <- seq(from = mu.vals[[1]][[j]][[1]], to = mu.vals[[1]][[j]][[2]], length=12)[-c(1,12)] +mean.base

   mu.vals[[1]][[j]] <- mu.eval-mean.base

   for (k in 1:10) {
   meanval[[j]][[1]][k] <- Gini(output$y.grid,w=compressed(mu.eval[k],cumsum(output$probs[[j]][[1]]),cumsum(output$probs[[j]][[2]]),output$y.grid)) - base
   meanval[[j]][[2]][k] <- Gini(output$y.grid,w=dispersed(mu.eval[k],cumsum(output$probs[[j]][[1]]),cumsum(output$probs[[j]][[2]]),output$y.grid)) - base
   }

   if (ginimax==TRUE) {
    meanval[[j]][[1]] <- max(meanval[[j]][[1]])
    meanval[[j]][[2]] <- min(meanval[[j]][[2]])
   }

  } else {
   meanval[[j]] <- Gini(output$y.grid,w=output$probs[[j]]) - base
  }

 }
 # END LOOP OVER VARIABLES

 meanval[[j+1]] <- Gini(output$y.grid,w=output$probs[[j+1]]) - base             # Full Policy Effect
 meanval[[j+2]] <- treat - base  # Observed change

 names(meanval) <- c(output$covname,"full","observed")

 out <- list(Gini=meanval, mean=mu.vals[[1]])


 # END: GINI
 ###############################################################################

 out

}



bootstrap.se <- function(output, stattype = "quantile", qseq=NA, mu.vals=NA, qc=NA...) {

  ## Two aux. functions that help with bootstrap calculations
   hilf1 <- function(object,variable,quant=1){
     if (is.list(object[[1]][[variable]])) {
      out <- c(object[[1]][[variable]][[1]][quant],object[[1]][[variable]][[1]][quant])
     } else {
     out <-  object[[1]][[variable]][quant]
     }
    out
   }

   hilf2 <- function(object){
    if (is.matrix(object)) {
     out <- apply(object,1,sd,na.rm = TRUE)
    } else {
   out <-  sd(object,na.rm = TRUE)
   }
   out
   }
   ############

   bs.data <- switch(stattype,
               mean = lapply(output$bs, fppe.mean.proc),
               quantile = lapply(output$bs, fppe.quantile.proc, qseq=qseq),
               var = lapply(output$bs, fppe.var.proc, mu.vals=mu.vals),
               gini = lapply(output$bs, fppe.gini.proc,mu.vals=mu.vals),
               varmax = lapply(output$bs, fppe.var.proc, mu.vals=mu.vals, varmax=TRUE),
               ginimax = lapply(output$bs, fppe.gini.proc, mu.vals=mu.vals, ginimax=TRUE),
               qc = lapply(output$bs, fppe.qc.proc, qc=qc))
   bs.serr <- list(NA)


 # special procedure for case of quantiles (list valued parameters)
 if (stattype %in% c("quantile","var","gini")) {

   # BEGIN: Loop over all variables
   for (j in 1:length(output$estimate$covname)) {

   if(output$estimate$covname[j] %in% output$estimate$discname) {
    bs.serr[[j]] <- list(NA)
    bs.serr[[j]][[1]] <-    rep(0, length(bs.data[[1]][[1]][[j]][[1]]))
    bs.serr[[j]][[2]] <-    rep(0, length(bs.data[[1]][[1]][[j]][[1]]))
    for (k in 1:length(bs.data[[1]][[1]][[j]][[1]])) {
     bs.serr[[j]][[1]][k] <-  hilf2(sapply(bs.data,hilf1,variable=j, quant=k))[1]
     bs.serr[[j]][[2]][k] <-  hilf2(sapply(bs.data,hilf1,variable=j, quant=k))[2]
    }
   } else {
    bs.serr[[j]] <- rep(0, length(bs.data[[1]][[1]][[j]]))
    for (k in 1:length(bs.data[[1]][[1]][[j]])) {
     bs.serr[[j]][k] <-  hilf2(sapply(bs.data,hilf1,variable=j, quant=k))
    }
   }
 }
 # END LOOP OVER VARIABLES

 bs.serr[[j+1]] <- rep(0,length(bs.data[[1]][[1]][[j]]))           # Full Policy Effect
 for (k in 1:length(bs.data[[1]][[1]][[j]])) {
     bs.serr[[j+1]][k] <-  hilf2(sapply(bs.data,hilf1,variable=j+1, quant=k))
    }

 bs.serr[[j+2]] <- rep(0,length(bs.data[[1]][[1]][[j]]))          # Observed change
 for (k in 1:length(bs.data[[1]][[1]][[j]])) {
     bs.serr[[j+2]][k] <-  hilf2(sapply(bs.data,hilf1,variable=j+2, quant=k))
    }

 names( bs.serr) <- c(output$estimate$covname,"full","observed")
 out <- bs.serr
 }


 if (stattype %in% c("mean","ginimax","varmax","qc")) {
   # BEGIN: Loop over all variables
   for (j in 1:length(output$estimate$covname)) {

   if(output$estimate$covname[j] %in% output$estimate$discname) {
     bs.serr[[j]]<- list(NA)
     bs.serr[[j]][[1]] <-  hilf2(sapply(bs.data,hilf1,variable=j))[1]
     bs.serr[[j]][[2]] <-  hilf2(sapply(bs.data,hilf1,variable=j))[2]

   } else {
     bs.serr[[j]] <-  hilf2(sapply(bs.data,hilf1,variable=j))

   }
 }
 # END LOOP OVER VARIABLES

  bs.serr[[j+1]] <-  hilf2(sapply(bs.data,hilf1,variable=j+1)) # Full Policy Effect
  bs.serr[[j+2]] <-  hilf2(sapply(bs.data,hilf1,variable=j+2)) # Observed change

  names( bs.serr) <- c(output$estimate$covname,"full","observed")
  out <- bs.serr

 }

out

}

################################################################################


fppe.quantile <- function(output, qseq=c(.1,.25,.5,.75,.9),...) {
  fppe <- fppe.quantile.proc(output$estimate,qseq=round(qseq,4))
  fppe.se <- NA

  if (length(output$bs)>1) {
   fppe.se <- bootstrap.se(output,stattype="quantile",qseq=round(qseq,4))
   }
  out <- list(estimate = fppe$quant, se = fppe.se, qseq=qseq)
  out }
  
fppe.mean <- function(output,...) {
  fppe <- fppe.mean.proc(output$estimate)
  fppe.se <- NA

  if (length(output$bs)>1) {
   fppe.se <- bootstrap.se(output,stattype="mean")
   }
  out <- list(estimate = fppe$mean, se = fppe.se)
  out }
  
  
fppe.var <- function(output,...) {
  mu.vals <- fppe.mean.proc(output$estimate)
  fppe <- fppe.var.proc(output$estimate,mu.vals=mu.vals,varmax=TRUE)
  fppe.se <- NA

  if (length(output$bs)>1) {
   fppe.se <- bootstrap.se(output,stattype="varmax",mu.vals=mu.vals)
   }
  out <- list(estimate = fppe$var, se = fppe.se,  mean=mu.vals[[1]] )
  out }
  
  
fppe.qc <- function(output,qc=c(.1,.9)) {
  fppe <- fppe.qc.proc(output$estimate,qc=qc)
  fppe.se <- NA

  if (length(output$bs)>1) {
   fppe.se <- bootstrap.se(output,stattype="qc",qc=qc)
   }
  out <- list(estimate = fppe$qc, se = fppe.se,  qc=qc )
  out }

fppe.gini <- function(output,...) {
  mu.vals <- fppe.mean.proc(output$estimate)
  fppe <- fppe.gini.proc(output$estimate,mu.vals=mu.vals,ginimax=TRUE)
  fppe.se <- NA

  if (length(output$bs)>1) {
   fppe.se <- bootstrap.se(output,stattype="ginimax",mu.vals=mu.vals)
   }
  out <- list(estimate = fppe$gini, se = fppe.se )
  out }






